"""
This module contains helper functions to analyze fear conditioning videos

@authors: Gergo Turi gt2253@cumc.columbia.edu"""

from os.path import join, basename
import struct
import subprocess
import time
from glob import glob

def ffii_locator(folder_path):
	""""
	finds FreezFrame (*.ffii) files within a folder

	Parameters:
	===========
	folder_path: str
		path like string where .ffii files are located

	Returns:
	========
	ffii_files: list
		list of .ffii files
	"""

	ffii_files  = []
	for name in glob(folder_path+'*.ffii'):
		ffii_files.append(name)
	return ffii_files

def ffii_converter(file_name, fps=4):
	"""
	converts Freezframe raw ffii binaries to .avi file. The output file
	will be located where the input was.

	Parameters:
	===========
	file_name = str
		path to a sigle file to process 
	fps = int or float
		the framerate the raw file will be converted with. 

	Return:
	=======
	Doesn't return anything. Writes the .ffii as avi.
	Note: could be used to save other types of videos as well by
	changing the extension in `of`.
	"""
  
	print(f'Converting {file_name}')

	#output file
	of = file_name[:-5]+'.avi'

	f = open(file_name,'rb')
	m = f.read(8)
	height, width= struct.unpack(">2I", m)
	rate = str(fps)

	cmdstr = ('ffmpeg', '-y', '-r', rate,\
	                    '-f', 'rawvideo',
	                    '-pix_fmt', 'gray',
	                    '-s', str(width)+"x"+str(height),
	                    '-i', '-',
	                    str(of))

	p = subprocess.Popen(cmdstr, stdin=subprocess.PIPE, shell=False)

	while True:
		img = f.read(width*height)
		p.stdin.write(img)
		m = f.read(8)
	if not m:
		break
	height, width = struct.unpack(">2I", m)

	print(f'File saved as {of}')
	f.close()              
	p.kill()


def ffii_counter(file_name):
	"""
	counts the number of frames in a raw Freezframe .ffii file.

	Parameters:
	===========
	file_name = str
		path to a sigle file to process 

	Returns:
	========
	frames: int
		number of frames
	"""
	frames = 0

	print(f'Counting frames in {file_name}')

	f = open(file_name,'rb')
	m = f.read(8)
	height, width= struct.unpack(">2I", m)

	while True:
		img = f.read(width*height)
		frames += 1
		m = f.read(8)
	if not m:
		break
	
	print(f'Number of frames: {frames}')
	                      
	f.close()

	return frames



"""
This module contains helper functions to analyze Y maza data

@authors: Gergo Turi gt2253@cumc.columbia.edu, Tingyi Lu tl3025@columbia.edu
"""
import os
import copy
import numpy as np
import pandas as pd
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.path as mplPath
from itertools import groupby
import matplotlib.patches as patches
import cv2
from google.colab.patches import cv2_imshow

def error_counter(data):
	"""
	Calculates the number of incorrect alternations

	Parameters:
	==========
	data : str
		the sequence of arm alternations.

	Return:
	=======
	errors: int
		number of alternation errors
	"""

	entries = list(data)
	# need to get rid of the first entries
	shortened_entries = entries[2:]
	errors = 0
	for i, letter in enumerate(shortened_entries):
		if shortened_entries[i] == (entries[i] or entries[i+1]):
			errors+=1
		
	return errors

def performance(data, errors):
	"""
	Calculates the Y maze performance of a mouse

	Parameters:
	===========
	data : str
		a single sequence of arm alternations.
	errors : int
		number of incorrect choices

	Return:
	=======
	performnce : float
	"""

	total_entries = len(list(data))
	potential_alternations = total_entries-2
	correct_choices = potential_alternations - errors
	performance = correct_choices / potential_alternations
	return performance

def plot_rois(rois, frame_file):
	'''
	Plots & marks ROIs that overlay the extracted frame from Y maze videos. Currently only works on Y mazes.
	
	Parameters:
	===========
	rois: nested list
		list of all ROIs; each sublist contains name of the ROI and the coodinates of its vertices
	frame_file: str
		a string that contains the directory of the extracted frame from the video being analyzed.
	
	Return:
	=======
	Returns nothing but the function outputs a plot.
	'''
	fig, ax = plt.subplots()
	for roi in rois:
		verts = np.array(roi[1])
		verts = np.append(verts, [verts[0]], axis=0)
		codes = [mplPath.Path.MOVETO, mplPath.Path.LINETO, mplPath.Path.LINETO,
		     mplPath.Path.LINETO, mplPath.Path.CLOSEPOLY]
		path = mplPath.Path(verts, codes) 
		patch = patches.PathPatch(path, facecolor=(1,1,1,0.5), lw=1, label=rois[0])
		ax.add_patch(patch)
		x = (verts[1][0]+verts[3][0])/2
		y = (verts[1][1]+verts[3][1])/2

		#Modify here for different font styles
		font = {'family' : 'sans',
		  'weight' : 'bold',
		  'size'   : 15,
		  'color': 'red'}
		plt.text(x, y, roi[0], **font)

	pic = cv2.imread(frame_file)
	plt.imshow(pic)
	ax.set_xlim(0, pic.shape[1])
	ax.set_ylim(0, pic.shape[0])
	ax.set_axis_off()
	plt.show()

def if_inside(roi, nose, tail_base):
	'''
	Checks whether the mouse is inside an ROI based on the coordinates of its nose tip and tail base.

	Parameters:
	===========
	roi: list of tuples
		the coordinates of all vertices of an ROI
	nose: tuple
		the coordinates of the nose of the mouse
	tail_base: tuple
		coordinates of the tail base of the mouse

	Return:
	=======
	inside: boolean
		True if the nose and tail_base are both inside the given ROI, False if otherwise
	'''
	inside=False

	roi = np.append(roi, [roi[0]], axis=0)
	poly_path = mplPath.Path(np.array(roi))
	nose = poly_path.contains_point(nose)
	tb = poly_path.contains_point(tail_base)

	if (nose==True) & (tb==True):
		inside=True

	return inside
	
def tracking_preprocess(tracking_data, likelihood_cutoff=0.8):
	'''
	Preprocesses the tracking data frame (contains coodinates of body parts of a mouse for every frame), drop unlikely data, and combine x, y coodinates

	Parameters:
	===========
	tracking_data: pandas data frame
		data frame of raw tracking data
	likelihood_cutoff: float
		default 0.8, threshold of likelihood, remove data points that have likelihood less than 0.8

	Return:
	=======
	df2: pandas data frame
		data frame of coordinates of nose and tail base to be used for processing arm entry later; each column contains coordinates in form of tuples
	'''
	df = tracking_data.copy()
	for part in df.columns.levels[0]:
		df_part = df[part]
		df = df.drop(df_part[df_part.likelihood < likelihood_cutoff].index)

	nose = df.loc[:, 'nose_tip']
	tb = df.loc[:, 'tail_base']

	nose_coord = [(x, y) for x, y in zip(nose.x, nose.y)]
	tb_coord = [(x, y) for x, y in zip(tb.x, tb.y)] 

	df2 = pd.DataFrame({'nose':nose_coord, 'tail_base':tb_coord})

	return df2

def generate_entry_sequence(df_body_coords, rois):
	'''
	generates the entry sequence of the mouse and grouped by three for better visibility

	Parameters:
	===========
	df_body_coords: pandas dataframe
	dataframe of nose and tail base coordinates generated from tracking_preprocess, obtained from function tracking_preprocess
	rois: nested list
	list of all ROIs; each sublist contains name of the ROI and the coodinates of its vertices

	Returns:
	========
	result_str: str
	the resulting arm entry sequence of the mouse, does not get rid of the first entry!
	'''
	df3 = pd.DataFrame()
	for arm in rois:
		df3[arm[0][-1]] = df_body_coords.apply(lambda x: if_inside(arm[1], x['nose'], x['tail_base']), axis=1)
	df4 = df3.stack().reset_index().drop(columns='level_0')
	df4.columns=['arm', 'is_in']
	df4 = df4.drop(df4[df4.is_in==False].index)
	result_seq = [x[0] for x in groupby(df4.arm.tolist())]

	result_str = (''.join([x for x in result_seq]))
	result_str = ' '.join([result_str[i:i+3] for i in range(0, len(result_str), 3)])

	return result_str



"""
This module contains helper functions to analyze Object tests, 
such as object location and novel object recognition 

@authors: Ayush Sinha -- as6430@columbia.edu;  Gergo Turi -- gt2253@cumc.columbia.edu
"""

def roi_analysis(rois,tracking_file_path,DLCscorer,bpt):
	
    """finding time spent by mouse in the traning dataset

    Parameters:
    ===========
    rois: 
    tracking_file_path: str
    a path string to the location of tracking
    DLCscorer: 
    bpt: str
    a body part which can be found in the DLC dataframe

    Returns:
    ========
    res: 

    """
    
    Dataframe = pd.read_hdf(tracking_file_path)
    
    #Filtering (check threshold)
    #DLCscorer = 'DLC_resnet50_OIT_OLTAug8shuffle1_1030000'
    #bpt='nose'
    
    #setting likelihood to 95% for confident locations.
    Dataframe = Dataframe[Dataframe[DLCscorer][bpt]['likelihood'] > 0.95]
    
    position = namedtuple('position', ['topleft', 'bottomright'])
    rois = {'Arena': position(rois[0][1][0],rois[0][1][2]),
            'Object1': position(rois[1][1][0],rois[1][1][2]),
            'Object2': position(rois[2][1][0],rois[2][1][2])} 

    xnose=Dataframe[DLCscorer][bpt]['x'].values
    ynose=Dataframe[DLCscorer][bpt]['y'].values
    
    bp_tracking = np.array((xnose,ynose))
    
    #assuming an FPS of 16
    fps = du.get_fps(vid_path)
    res = tr.get_timeinrois_stats(bp_tracking.T, rois, 16, False, False)
    
    return res



import numpy as np
import pandas as pd
import os
import pickle
import sys
import tables
from scipy.spatial import distance
import warnings

def calc_distance_between_points_in_a_vector_2d(v1):
    '''calc_distance_between_points_in_a_vector_2d [for each consecutive pair of points, p1-p2, in a vector, get euclidian distance]
    This function can be used to calculate the velocity in pixel/frame from tracking data (X,Y coordinates)
    
    Arguments:
        v1 {[np.array]} -- [2d array, X,Y position at various timepoints]
    
    Raises:
        ValueError
    
    Returns:
        [np.array] -- [1d array with distance at each timepoint]
    >>> v1 = [0, 10, 25, 50, 100]
    >>> d = calc_distance_between_points_in_a_vector_2d(v1)
    '''
    # Check data format
    if isinstance(v1, dict) or not np.any(v1) or v1 is None:
            raise ValueError(
                'Feature not implemented: cant handle with data format passed to this function')

    # If pandas series were passed, try to get numpy arrays
    try:
        v1, v2 = v1.values, v2.values
    except:  # all good
        pass
    # loop over each pair of points and extract distances
    dist = []
    for n, pos in enumerate(v1):
        # Get a pair of points
        if n == 0:  # get the position at time 0, velocity is 0
            p0 = pos
            dist.append(0)
        else:
            p1 = pos  # get position at current frame

            # Calc distance
            dist.append(np.abs(distance.euclidean(p0, p1)))

            # Prepare for next iteration, current position becomes the old one and repeat
            p0 = p1

    return np.array(dist)
    

def selector(a):
    if a[1] == 1:
        return 1
    elif a[2] == 1:
        return 2
    elif a[0] == 1:
        return 0
    return 0
    
def get_roi_at_each_frame(bp_data, rois, check_inroi):
    """
    Given position data for a bodypart and the position of a list of rois, this function calculates which roi is
    the closest to the bodypart at each frame
    :param bp_data: numpy array: [nframes, 3] -> X,Y,Speed position of bodypart at each frame
                    [as extracted by DeepLabCut] --> df.bodypart.values. 
    :param rois: dictionary with the position of each roi. The position is stored in a named tuple with the location of
                    two points defyining the roi: topleft(X,Y) and bottomright(X,Y).
    :param check_inroi: boolean, default True. If true only counts frames in which the tracked point is inside of a ROI. 
                Otherwise at each frame it counts the closest ROI.
    :return: tuple, closest roi to the bodypart at each frame
    """

    def sort_roi_points(roi):
        return np.sort([roi.topleft[0], roi.bottomright[0]]), np.sort([roi.topleft[1], roi.bottomright[1]])

    if not isinstance(rois, dict): raise ValueError('rois locations should be passed as a dictionary')

    if not isinstance(bp_data, np.ndarray):
        if not isinstance(bp_data, tuple): raise ValueError('Unrecognised data format for bp tracking data')
        else:
            pos = np.zeros((len(bp_data.x), 2))
            pos[:, 0], pos[:, 1] = bp_data.x, bp_data.y
            bp_data = pos
            
    data_length = bp_data.shape[0]
    distances = np.zeros((data_length, len(rois)))
    for idx, pos in enumerate(rois.values()):
        cnt = np.tile(list(pos.topleft) + list(pos.bottomright), data_length).reshape((data_length, 4))

        dist = (cnt[:, 0] < bp_data[:, 0]) & (bp_data[:, 0]  < cnt[:,2]) & (cnt[:, 1] < bp_data[:, 1]) & (bp_data[:,1] < cnt[:, 3])
        distances[:, idx] = dist

    roi_names = list(rois.keys())
    # Get the center of each roi
    #     centers = []
    #     for points in rois.values():
    #         center_x = (points.topleft[0] + points.bottomright[0]) / 2
    #         center_y = (points.topleft[1] + points.bottomright[1]) / 2
    #         center = np.asarray([center_x, center_y])
    #         centers.append(center)

    #     roi_names = list(rois.keys())

    # Calc distance to each roi for each frame
    #     data_length = bp_data.shape[0]
    #     distances = np.zeros((data_length, len(centers)))
    #     for idx, center in enumerate(centers):
    #         cnt = np.tile(center, data_length).reshape((data_length, 2))
        
    #         dist = np.hypot(np.subtract(cnt[:, 0], bp_data[:, 0]), np.subtract(cnt[:, 1], bp_data[:, 1]))
    #         distances[:, idx] = dist

    # Get which roi is closest at each frame
    #     sel_rois = np.argmin(distances, 1)
    #     roi_at_each_frame = tuple([roi_names[x] for x in sel_rois])

    # Check if the tracked point is actually in the closest ROI

    sel_rois = np.apply_along_axis(selector, 1, distances)
    roi_at_each_frame = tuple([roi_names[x] for x in sel_rois])

    if not check_inroi: 
        cleaned_rois = []
        for i, roi in enumerate(roi_at_each_frame):
            x,y = bp_data[i, 0], bp_data[i, 1]
            X, Y = sort_roi_points(rois[roi]) # get x,y coordinates of roi points
            if not X[0] <= x <= X[1] or not Y[0] <= y <= Y[1]:
                cleaned_rois.append('none')
            else:
                cleaned_rois.append(roi)
        #print(cleaned_rois)
        return cleaned_rois
    else:
        warnings.warn("Warning: you've set check_inroi=False, so data reflect which ROI is closest even if tracked point is not in any given ROI.")
        return roi_at_each_frame
    

def get_time_in_rois_stats(data, rois, fps=None, returndf=False, check_inroi=True):
    """
    Quantify number of times the animal enters a roi, cumulative number of frames spend there, cumulative time in seconds
    spent in the roi and average velocity while in the roi.
    In which roi the mouse is at a given frame is determined with --> get_roi_at_each_frame()
    Quantify the ammount of time in each  roi and the avg stay in each roi
    :param data: trackind data is a numpy array with shape (n_frames, 3) with data for X,Y position and Speed. If [n_frames, 2]
                array is passed, speed is calculated automatically.
    :param rois: dictionary with the position of each roi. The position is stored in a named tuple with the location of
                two points defyining the roi: topleft(X,Y) and bottomright(X,Y).
    :param fps: framerate at which video was acquired
    :param returndf: boolean, default False. If true data are returned as a DataFrame instead of dict.
    :param check_inroi: boolean, default True. If true only counts frames in which the tracked point is inside of a ROI. 
                Otherwise at each frame it counts the closest ROI.
    :return: dictionary or dataframe
    # Testing
    >>> position = namedtuple('position', ['topleft', 'bottomright'])
    >>> rois = {'middle': position((300, 400), (500, 800))}
    >>> data = np.zeros((23188, 3))
    >>> res = get_timeinrois_stats(data, rois, fps=30)
    >>> print(res)
    """

    def get_indexes(lst, match):
        return np.asarray([i for i, x in enumerate(lst) if x == match])

    # Check arguments
    if data.shape[1] == 2:  # only X and Y tracking data passed, calculate speed
        speed = calc_distance_between_points_in_a_vector_2d(data)
        data = np.hstack((data, speed.reshape((len(speed), 1))))

    elif data.shape[1] != 3:
        raise ValueError("Tracking data should be passed as either an Nx2 or Nx3 array."\
         "Tracking data shape was: {}. Maybe you forgot to transpose the data?".format(data.shape))

    roi_names = [k.lower() for k in list(rois.keys())]
    if "none" in roi_names:
        raise ValueError("No roi can have name 'none'," \
            " that's reserved for the code to use, please use a different name for your rois.")

    if "tot" in roi_names:
        raise ValueError("No roi can have name 'tot'," \
            "that's reserved for the code to use, please use a different name for your rois.")

    # get roi at each frame of data
    data_rois = get_roi_at_each_frame(data, rois, check_inroi)
    data_time_inrois = {name: data_rois.count(name) for name in set(data_rois)}  # total time (frames) in each roi

    # number of enters in each roi
    transitions = [n for i, n in enumerate(list(data_rois)) if i == 0 or n != list(data_rois)[i - 1]]
    transitions_count = {name: transitions.count(name) for name in transitions}

    # avg time spend in each roi (frames)
    avg_time_in_roi = {transits[0]: time / transits[1]
                       for transits, time in zip(transitions_count.items(), data_time_inrois.values())}

    # avg time spend in each roi (seconds)
    if fps is not None:
        data_time_inrois_sec = {name: t / fps for name, t in data_time_inrois.items()}
        avg_time_in_roi_sec = {name: t / fps for name, t in avg_time_in_roi.items()}
    else:
        data_time_inrois_sec, avg_time_in_roi_sec = None, None

    # get avg velocity in each roi
    avg_vel_per_roi = {}
    for name in set(data_rois):
        indexes = get_indexes(data_rois, name)
        vels = data[indexes, 2]
        avg_vel_per_roi[name] = np.average(np.asarray(vels))

    # get cumulative
    transitions_count['tot'] = np.sum(list(transitions_count.values()))
    data_time_inrois['tot'] = np.sum(list(data_time_inrois.values()))
    data_time_inrois_sec['tot'] = np.sum(list(data_time_inrois_sec.values()))
    avg_time_in_roi['tot'] = np.sum(list(avg_time_in_roi.values()))
    avg_time_in_roi_sec['tot'] = np.sum(list(avg_time_in_roi_sec.values()))
    avg_vel_per_roi['tot'] = np.sum(list(avg_vel_per_roi.values()))

    if returndf:
        roinames = sorted(list(data_time_inrois.keys()))
        results = pd.DataFrame.from_dict({
                    "ROI_name": roinames, 
                    "transitions_per_roi": [transitions_count[r] for r in roinames],
                    "cumulative_time_in_roi": [data_time_inrois[r] for r in roinames],
                    "cumulative_time_in_roi_sec": [data_time_inrois_sec[r] for r in roinames],
                    "avg_time_in_roi": [avg_time_in_roi[r] for r in roinames],
                    "avg_time_in_roi_sec": [avg_time_in_roi_sec[r] for r in roinames],
                    "avg_vel_in_roi": [avg_vel_per_roi[r] for r in roinames],
                    })
    else:
        results = dict(transitions_per_roi=transitions_count,
                cumulative_time_in_roi=data_time_inrois,
                cumulative_time_in_roi_sec=data_time_inrois_sec,
                avg_time_in_roi=avg_time_in_roi,
                avg_time_in_roi_sec=avg_time_in_roi_sec,
                avg_vel_in_roi=avg_vel_per_roi)

    return results


import numpy as np
import pandas as pd
import os
import pickle
import sys
import tables
from scipy.spatial import distance
import warnings
import time_in_rois as tr
from collections import namedtuple

def roi_analysis(rois,tracking_file_path):
    #currently finding time spent by mouse in the traning dataset
    
    Dataframe = pd.read_hdf(tracking_file_path)
    
    #Filtering (check threshold)
    DLCscorer = 'DLC_resnet50_OIT_OLTAug8shuffle1_1030000'
    bpt='nose'

    Dataframe = Dataframe[Dataframe[DLCscorer][bpt]['likelihood'] > 0.95]
    
    position = namedtuple('position', ['topleft', 'bottomright'])
    rois = {'Arena': position(rois[0][1][0],rois[0][1][2]),
            'Object1_Day1': position(rois[1][1][0],rois[1][1][2]),
            'Object2_Day1': position(rois[2][1][0],rois[2][1][2])} 
            
    xnose=Dataframe[DLCscorer][bpt]['x'].values
    ynose=Dataframe[DLCscorer][bpt]['y'].values
    
    bp_tracking = np.array((xnose,ynose))
    res = tr.get_timeinrois_stats(bp_tracking.T, rois, 10, False, False)
    
    return res